import logging
import time
import os


def get_logger(name, log_level=logging.DEBUG):
    """一般日志设置所采用的方式"""
    log = logging.getLogger(name)
    log.setLevel(log_level)
    directory = os.path.join('log', time.strftime('%Y%m%d-%H%M%S'))
    if not os.path.isdir(directory):
        os.makedirs(directory, exist_ok=True)
    filename = os.path.join(directory, '{}.log'.format(name))
    fh = logging.FileHandler(filename)
    formatter = logging.Formatter('%(asctime)s - %(processName)s - '
                                  '%(threadName)s - %(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    log.addHandler(fh)
    return log



"""paralleller namespace."""

from .jl import JoblibPool
from .mpi import MPIPool
from .serial import SerialPool
from .multiprocessing import MultiPool
from importlib_metadata import PackageNotFoundError, version

__author__ = "Rontomai"
__email__ = "rontomai@gmail.com"

# Used to automatically set version number from github actions
# as well as not break when being tested locally
try:
    __version__ = version(__package__)  # type: ignore
except PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0"


import sys
from log import get_logger
log = get_logger(__name__)
_VERBOSE = 5


def choose_pool(mpi=False, processes=1, **kwargs):
    """
    Choose between the different pools given options from, e.g., argparse.
    Parameters
    ----------
    mpi : bool, optional
        Use the MPI processing pool, :class:`~schwimmbad.mpi.MPIPool`. By
        default, ``False``, will use the :class:`~schwimmbad.serial.SerialPool`.
    processes : int, optional
        Use the multiprocessing pool,
        :class:`~schwimmbad.multiprocessing.MultiPool`, with this number of
        processes. By default, ``processes=1``, will use the
        :class:`~schwimmbad.serial.SerialPool`.
    **kwargs
        Any additional kwargs are passed in to the pool class initializer
        selected by the arguments.
    """

    if mpi:
        if not MPIPool.enabled():
            raise SystemError("Tried to run with MPI but MPIPool not enabled.")

        pool = MPIPool(**kwargs)
        log.info("Running with MPI on {0} cores".format(pool.size))
        return pool

    elif processes != 1 and MultiPool.enabled():
        log.info("Running with MultiPool on {0} cores".format(processes))
        return MultiPool(processes=processes, **kwargs)

    else:
        log.info("Running with SerialPool")
        return SerialPool(**kwargs)



try:
    from joblib import Parallel, delayed
except ImportError:
    Parallel = None

from .pool import BasePool

__all__ = ["JoblibPool"]


class JoblibPool(BasePool):
    """A processing pool that distributes tasks using ``joblib.Parallel``.

    This pool processes parallel tasks using the ``Parallel`` module of the
    `joblib <https://pythonhosted.org/joblib>`_ library. All arguments and
    keyword arguments are passed directly to the ``__init__`` method of the
    `Parallel object provided by joblib
    <https://pythonhosted.org/joblib/parallel.html#parallel-reference-documentation>`_.

    """

    def __init__(self, *args, **kwargs):
        if Parallel is None:
            raise ImportError("joblib is required to use the JoblibPool")
        self.args = args
        self.kwargs = kwargs
        self.size = 0
        self.rank = 0

    @staticmethod
    def enabled():
        return Parallel is not None

    def map(self, func, iterable, callback=None):
        dfunc = delayed(func)
        res = Parallel(*(self.args), **(self.kwargs))(
            dfunc(a) for a in iterable
        )
        return self._call_callback(callback, res)



"""Main module."""
from itertools import starmap 
from typing import TYPE_CHECKING 


if TYPE_CHECKING:
    from typing import Optional, Callable, Dict 
    from collections.abc import Iterable 


def _map_parallel_multiprocessing(
    f: Callable,
    *args: Iterable,
    processes: Optional[int] = None,
    return_results: bool = True
) -> list:
    from concurrent.futures import ProcessPoolExecutor 

    with ProcessPoolExecutor(max_workers=processes) as process_pool_executor:
        res = process_pool_executor.map(f, *args)
        if return_results:
            return list(res)
        else:
            return []


def _map_parallel_multithreading(
    f: Callable, 
    *args: Iterable,
    processes: Optional[int] = None,
    return_results: bool = True
) -> list:
    from concurrent.futures import ThreadPoolExecutor

    with ThreadPoolExecutor(max_workers=processes) as thread_pool_executor:
        res = thread_pool_executor.map(f, *args)
        if return_results:
            return list(res)
        else:
            return []


def _map_parallel_dask(
    f: Callable,
    *args: Iterable,
    processes: Optional[int] = None,
    return_results: bool = True
) -> list:
    from dask.distributed import Client 
    from dask.distributed import LocalCluster 

    cluster = LocalCluster(max_workers=processes, dashboard=None)
    client = Client(cluster)
    if return_results:
        return [future.result() for future in client.map(f, *args)] 
    else:
        for future in client.map(f, *args):
            future.result()
        return []


def _map_parallel_mpi(
    f: Callable,
    args: Iterable[Iterable],
    return_results: bool = True,
    **kwargs
) -> list:
    from mpi4py.futures import MPIPoolExecutor

    with MPIPoolExecutor() as mpi_pool_executor:
        res = mpi_pool_executor.map(f, args)
        if return_results:
            return list(res)
        else:
            return []


def _starmap_parallel_mpi(
    f: Callable, 
    args: Iterable[Iterable],
    return_results: bool = True,
    **kwargs  
) -> list:
    from mpi4py.futures import MPIPoolExecutor

    with MPIPoolExecutor() as mpi_pool_executor:
        res = mpi_pool_executor.starmap(f, args) 
        if return_results:
            return list(res)
        else:
            return []


def _starmap_parallel_mpi_simple(
    f: Callable, 
    args: Iterable[Iterable],
    return_results: bool = True,
    **kwargs,
) -> list:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    size = comm.Get_size()
    rank = comm.Get_rank()

    args_list = list(args)
    if args_list:
        n = len(args_list)
        start = (rank * n) // size 
        end = (( rank + 1) * n) // size 
        local_args = args_list[start:end]
        res = list(starmap(f, local_args))

        if return_results:
            res = comm.gather(res, root=0)
            if rank == 0:
                return sum(res, [])
    return []


_map_parallel_func: 'Dict[str, Callable]' = {
    'multiprocessing': _map_parallel_multiprocessing,
    'multithreading': _map_parallel_multithreading,
    'dask': _map_parallel_dask,
    'mpi': _map_parallel_mpi,
}


_starmap_parallel_func: 'Dict[str, Callable]' = {
    'mpi': _starmap_parallel_mpi,
    'mpi_simple': _starmap_parallel_mpi_simple,
}


def map_parallel(
    f: 'Callable',
    *args: 'Iterable',
    processes: 'Optional[int]' = None,
    mode: 'str' = 'multiprocessing',
    return_results: 'bool' = True,
) -> 'list':
    '''equiv to `map(f, *args)` but in parallel
    :param str mode: backend for parallelization
        - multiprocessing: using multiprocessing from standard library
        - multithreading: using multithreading from standard library
        - dask: using dask.distributed
        - mpi: using mpi4py.futures. May not work depending on your MPI vendor
        - mpi_simple: using mpi4py with simple scheduling that divides works into equal chunks
        - serial: using map
    :param int processes: no. of parallel processes
        (in the case of mpi, it is determined by mpiexec/mpirun args)
    :param bool return_results: (Only affects mode == 'mpi_simple') if True, return results in rank 0.
    '''
    if processes is None or processes > 1:
        if mode in _map_parallel_func:
            return _map_parallel_func[mode](f, *args, processes=processes, return_results=return_results)
        elif mode in _starmap_parallel_func:
            return _starmap_parallel_func[mode](f, zip(*args), processes=processes, return_results=return_results)
    return list(map(f, *args))


def starmap_parallel(
    f: 'Callable',
    args: 'Iterable[Iterable]',
    processes: 'Optional[int]' = None,
    mode: 'str' = 'multiprocessing',
    return_results: 'bool' = True,
) -> 'list':
    '''equiv to `starmap(f, args)` but in parallel
    See docstring from :func:`~map_parallel.map_parallel`
    '''
    if processes is None or processes > 1:
        if mode in _starmap_parallel_func:
            return _starmap_parallel_func[mode](f, args, processes=processes, return_results=return_results)
        elif mode in _map_parallel_func:
            return _map_parallel_func[mode](f, *zip(*args), processes=processes, return_results=return_results)
    return list(starmap(f, args))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['paralleller']

package_data = \
{'': ['*']}

install_requires = \
['importlib_metadata>=3.4.0,<4.0.0']

setup_kwargs = {
    'name': 'paralleller',
    'version': '0.0.0',
    'description': 'a interface used to parallel processing pools.',
    'long_description': '# paralleller 介绍\n\n\n[![PyPI version](https://badge.fury.io/py/paralleller.svg)](https://badge.fury.io/py/paralleller)\n![versions](https://img.shields.io/pypi/pyversions/paralleller.svg)\n[![GitHub license](https://img.shields.io/github/license/mgancita/paralleller.svg)](https://github.com/mgancita/paralleller/blob/main/LICENSE)\n\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n\n用于进行并行处理的接口库。\n\n\n- 开源许可: MIT\n- 文档: https://llango.github.io/paralleller.\n\n\n## 特征\n\n* TODO\n\n\n## 项目布局\n\n    main.py    # 项目主文件，服务使用它启动。 uvicorn main:app --reload\n    .env       # 环境变量文件\n    requirements.txt    # 依赖文件\n    readme.md       # 简单说明文件\n    test.db         # 生成的数据库文件\n    mkdocs.yml      # mkdocs文档构建配置文件\n    LICENSE         # 开源许可文件\n    Procfile        # 容器用来执行用户设定的命令\n    Makefile        # 用来启动docker-compose 文件\n    .dockerignore   # docker打包上传到容器里过滤文件\n    .gitignore      # 版本控制过滤上传文件\n    docker-compose.yaml # docker-compose配置启动文件\n    Dockerfile      # Docker文件\n    core/\n        blog.py     # 与数据库进行博客交互文件\n        user.py     # 与数据库进行用户交互文件\n        __init__.py # 声明为包所需文件\n    database/\n        config.py   # 数据库连接配置文件\n        __init__.py # 声明为包所需文件\n    docs/\n        index.md    # 文档首页\n        database_config.md      # 对应database/config.py 文件的讲解\n    models/\n        models.py   # 用来建立数据库表映射的类文件\n        __init__.py # 声明为包所需文件\n    route/\n        auth.py     # 登录认证接口文件\n        blog.py     # 博客接口文件\n        user.py     # 用户接口文件\n        __init__.py # 声明为包所需文件\n    schema/\n        schemas.py  # 声明类型文件\n        __init__.py # 声明为包所需文件\n    static/\n        css/\n            index.css   # 样式文件\n        js/\n            main.js     # 脚步\n        images/         # 图片文件夹\n    templates/\n        index.html      # 模板文件\n    test/               # 测试目录\n    utils/\n        hash.py         # 密码处理文件\n        oa2.py          # 认证处理文件\n        token.py        # 指令处理文件\n        __init__.py     # 声明为包所需文件  \n\n\n## 项目启动\n    1. 进行该项目中，建立虚拟环境\n    ```\n    virtualenv venv\n    source  venv/bin/activate\n    ```\n    2. 安装依赖\n    ```\n    pip install -r requirements.txt\n    ```\n    3. 直接使用如下命令启动:\n        ``` \n        uvicorn main:app --realod \n        ```\n    \n## 文档启动\n\n```\nmkdocs serve --livereload -t shadocs -a localhost:8080\n```\n\n\n## 制作\n\n\n该包使用 [Cookiecutter](https://github.com/audreyr/cookiecutter) 和 [`llango/cookiecutter-mkdoc-shapackage`](https://github.com/llango/cookiecutter-mkdoc-shapackage/) 项目模版创建。\n',
    'author': 'Rontomai',
    'author_email': 'rontomai@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/llango/paralleller',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
